{-# LANGUAGE DeriveGeneric #-}
{-# LANGUAGE DerivingVia #-}
module Ubc.Parse.Syntax.File
( File(..)
, parse
)
where

import GHC.Generics (Generic, Generically(..))

import Control.Monad ((<$!>))
import Data.Functor ((<&>))

import Text.Parsec (choice, ParsecT, many)

import Ubc.Parse.Syntax (Transformer)

import {-# SOURCE #-} Ubc.Parse.Syntax.Import (Import)
import Ubc.Parse.Syntax.Struct (Struct)
import Ubc.Parse.Syntax.Function (Function)
import Ubc.Parse.Syntax.Statement (Statement)
import Ubc.Parse.Syntax.Enumeration (Enumeration)

import {-# SOURCE #-} qualified Ubc.Parse.Syntax.Import as Import
import qualified Ubc.Parse.Syntax.Struct as Struct
import qualified Ubc.Parse.Syntax.Function as Function
import qualified Ubc.Parse.Syntax.Statement as Statement
import qualified Ubc.Parse.Syntax.Enumeration as Enumeration
import qualified Text.Parsec as Parsec

data File = File
        { name :: String
        , body :: FileBody
        }
        deriving (Show)

data FileBody = FileBody
        { imports :: [Import]
        , structs :: [Struct]
        , functions :: [Function]
        , statements :: [Statement]
        , enumerations :: [Enumeration]
        }
        deriving stock (Show, Generic)
        deriving (Semigroup, Monoid) via Generically FileBody
-- dont use `deriving ... via FileBody` because that leads to a loop, somehow

parse :: ParsecT String u Transformer File
parse = do
        fileName <- Parsec.sourceName <$!> Parsec.getPosition
        fileBody <- mconcat <$!> many fileMember
        pure $ File fileName fileBody

fileMember :: ParsecT String u Transformer FileBody
fileMember = choice
                [ Struct.parse      <&> \s -> mempty { structs = [s] }
                , Import.parse      <&> \i -> mempty { imports = [i] }
                , Function.parse    <&> \f -> mempty { functions = [f] }
                , Statement.parse   <&> \s -> mempty { statements = [s] }
                , Enumeration.parse <&> \e -> mempty { enumerations = [e] }
                ]
